using System;
using System.Drawing;
using System.Windows.Forms;

using Emgu.CV;
using Emgu.CV.CvEnum;
using Emgu.CV.Structure;
using Emgu.CV.VideoSurveillance;

namespace TestProject
{
    /// <summary>
    ///  
    /// </summary>
    public partial class MainForm : Form
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// ĸó
        /// </summary>
        private Capture capture;

        /// <summary>
        ///  丮
        /// </summary>
        private MotionHistory motionHistory;

        /// <summary>
        /// BGFG 
        /// </summary>
        private IBGFGDetector<Bgr> bgfgDetector;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region  - MainForm()

        /// <summary>
        /// 
        /// </summary>
        public MainForm()
        {
            InitializeComponent();

            if(this.capture == null)
            {
                try
                {
                    this.capture = new Capture();
                }
                catch(Exception exception)
                {
                    MessageBox.Show(exception.Message);
                }
            }

            if(this.capture != null)
            {
                this.motionHistory = new MotionHistory
                (
                    1.0,  //  丮  ð ( )
                    0.05, // cvCalcMotionGradient maxDelta ( )
                    0.5   // cvCalcMotionGradient minDelta ( )
                );

                this.capture.ImageGrabbed += capture_ImageGrabbed;

                this.capture.Start();
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Private
        //////////////////////////////////////////////////////////////////////////////// Event

        #region  ݴ  óϱ - Form_FormClosed(sender, e)

        /// <summary>
        ///  ݴ  óϱ
        /// </summary>
        /// <param name="sender">̺Ʈ ߻</param>
        /// <param name="e">̺Ʈ </param>
        private void Form_FormClosed(object sender, FormClosedEventArgs e)
        {
            this.capture.Stop();
        }

        #endregion
        #region ĸó ̹ GRAB óϱ - capture_ImageGrabbed(sender, e)

        /// <summary>
        /// ĸó ̹ GRAB óϱ
        /// </summary>
        /// <param name="sender">̺Ʈ ߻</param>
        /// <param name="e">̺Ʈ </param>
        private void capture_ImageGrabbed(object sender, EventArgs e)
        {
            using(Image<Bgr, Byte> captureImage = this.capture.RetrieveBgrFrame())
            {
                using(MemStorage storage = new MemStorage())
                {
                    if(this.bgfgDetector == null)
                    {
                        this.bgfgDetector = new FGDetector<Bgr>(FORGROUND_DETECTOR_TYPE.FGD);
                    }

                    this.bgfgDetector.Update(captureImage);

                    this.captureImageBox.Image = captureImage;

                    this.motionHistory.Update(this.bgfgDetector.ForegroundMask);

                    this.forgroundMaskImageBox.Image = this.bgfgDetector.ForegroundMask;

                    #region  ũ 纻 ϰ  Ѵ.

                    double[] minimumValueArray;
                    double[] maximumValueArray;
                    Point[]  minimumPointArray;
                    Point[]  maximumPointArray;

                    this.motionHistory.Mask.MinMax(out minimumValueArray, out maximumValueArray, out minimumPointArray, out maximumPointArray);

                    Image<Gray, Byte> motionMaskImage = this.motionHistory.Mask.Mul(255.0 / maximumValueArray[0]);

                    #endregion
                    #region  ̹ Ѵ.

                    Image<Bgr, Byte> motionImage = new Image<Bgr, byte>(motionMaskImage.Size);

                    motionImage[0] = motionMaskImage;

                    #endregion

                    //   ϴ Ѱ ̼    δ.
                    double minimumArea = 100;

                    storage.Clear();

                    Seq<MCvConnectedComp> motionComponentSequence = this.motionHistory.GetMotionComponents(storage);

                    foreach(MCvConnectedComp motionComponent in motionComponentSequence)
                    {
                        if(motionComponent.area < minimumArea)
                        {
                            continue;
                        }

                        // Ư    ȼ īƮ ã´.
                        double angle;
                        double motionPixelCount;

                        this.motionHistory.MotionInfo(motionComponent.rect, out angle, out motionPixelCount);

                        if(motionPixelCount < motionComponent.area * 0.05)
                        {
                            continue;
                        }

                        DrawMotion(motionImage, motionComponent.rect, angle, new Bgr(Color.Red));
                    }

                    // ü   ã ׸.
                    double overallAngle;
                    double overallMotionPixelCount;

                    this.motionHistory.MotionInfo(motionMaskImage.ROI, out overallAngle, out overallMotionPixelCount);

                    DrawMotion(motionImage, motionMaskImage.ROI, overallAngle, new Bgr(Color.Green));

                    UpdateMessage
                    (
                        string.Format
                        (
                            "ü Ž  : {0};  ȼ īƮ : {1}",
                            motionComponentSequence.Total,
                            overallMotionPixelCount
                        )
                    );

                    this.motionImageBox.Image = motionImage;
                }
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Function

        #region ޽ ϱ - UpdateMessage(message)

        /// <summary>
        /// ޽ ϱ
        /// </summary>
        /// <param name="message">޽</param>
        private void UpdateMessage(string message)
        {
            if(InvokeRequired && !IsDisposed)
            {
                Invoke((Action<string>)UpdateMessage, message);
            }
            else
            {
                this.message.Text = message;
            }
        }

        #endregion
        #region  ׸ - DrawMotion(image, motionRectangle, angle, color)

        /// <summary>
        ///  ׸
        /// </summary>
        /// <param name="image">̹</param>
        /// <param name="motionRectangle"> 簢</param>
        /// <param name="angle"></param>
        /// <param name="color"></param>
        private static void DrawMotion(Image<Bgr, Byte> image, Rectangle motionRectangle, double angle, Bgr color)
        {
            float circleRadius = (motionRectangle.Width + motionRectangle.Height) >> 2;

            Point centerPoint = new Point
            (
                motionRectangle.X + motionRectangle.Width  >> 1,
                motionRectangle.Y + motionRectangle.Height >> 1
            );

            CircleF circle = new CircleF(centerPoint, circleRadius);

            int directionX = (int)(Math.Cos(angle * (Math.PI / 180.0)) * circleRadius);
            int directionY = (int)(Math.Sin(angle * (Math.PI / 180.0)) * circleRadius);

            Point pointOnCircle = new Point(centerPoint.X + directionX, centerPoint.Y - directionY);

            LineSegment2D line = new LineSegment2D(centerPoint, pointOnCircle);

            image.Draw(circle, color, 1);

            image.Draw(line, color, 2);
        }

        #endregion
    }
}